@@ -14,26 +14,22 @@ use constant {
 	# Lowest compat level that does *not* cause deprecation
 	# warnings
 	'LOWEST_NON_DEPRECATED_COMPAT_LEVEL' => 9,
-	# Highest "open-beta" compat level.  Remember to notify
-	# debian-devel@l.d.o before bumping this.
-	'BETA_TESTER_COMPAT' => 10,
 	# Highest compat level permitted
-	'MAX_COMPAT_LEVEL' => 11,
+	'MAX_COMPAT_LEVEL' => 12,
 	# Magic value for xargs
 	'XARGS_INSERT_PARAMS_HERE' => \'<INSERT-HERE>', #'# Hi emacs.
 	# Magic value for debhelper tools to request "current version"
 	'DH_BUILTIN_VERSION' => \'<DH_LIB_VERSION>', #'# Hi emacs.
+	# Default Package-Type / extension (must be aligned with dpkg)
+	'DEFAULT_PACKAGE_TYPE' => 'deb',
 };
 
-my %NAMED_COMPAT_LEVELS = (
-	# The bleeding-edge compat level is deliberately not documented.
-	# You are welcome to use it, but please subscribe to the git
-	# commit mails if you do.  There is no heads up on changes for
-	# bleeding-edge testers as it is mainly intended for debhelper
-	# developers.
-	'bleeding-edge-tester' => MAX_COMPAT_LEVEL,
-	'beta-tester'          => BETA_TESTER_COMPAT,
-);
+use constant {
+	# Package-Type / extension for dbgsym packages
+	# TODO: Find a way to determine this automatically from the vendor
+	#  - blocked by Dpkg::Vendor having a rather high load time (for debhelper)
+	'DBGSYM_PACKAGE_TYPE' => DEFAULT_PACKAGE_TYPE,
+};
 
 use Errno qw(ENOENT);
 use Exporter qw(import);
@@ -64,7 +60,10 @@ our (@EXPORT, %dh);
 	    &glob_expand_error_handler_warn_and_discard &glob_expand
 	    &glob_expand_error_handler_silently_ignore DH_BUILTIN_VERSION
 	    &print_and_complex_doit &default_sourcedir &qx_cmd
-	    &compute_doc_main_package
+	    &compute_doc_main_package &is_so_or_exec_elf_file &hostarch
+	    &assert_opt_is_known_package &dbgsym_tmpdir &find_hardlinks
+	    &should_use_root &gain_root_cmd DEFAULT_PACKAGE_TYPE
+	    DBGSYM_PACKAGE_TYPE
 );
 
 # The Makefile changes this if debhelper is installed in a PREFIX.
@@ -73,6 +72,13 @@ my $prefix="/usr";
 my $MAX_PROCS = get_buildoption("parallel") || 1;
 my $DH_TOOL_VERSION;
 
+our $PKGNAME_REGEX = qr/[a-z0-9][-+\.a-z0-9]+/o;
+our $PKGVERSION_REGEX = qr/
+                 (?: \d+ : )?                # Optional epoch
+                 [0-9][0-9A-Za-z.+:~]*       # Upstream version (with no hyphens)
+                 (?: - [0-9A-Za-z.+:~]+ )*   # Optional debian revision (+ upstreams versions with hyphens)
+                          /xoa;
+
 sub init {
 	my %params=@_;
 
@@ -108,7 +114,7 @@ sub init {
 	if ((defined $ENV{DH_OPTIONS} && length $ENV{DH_OPTIONS}) ||
  	    (defined $ENV{DH_INTERNAL_OPTIONS} && length $ENV{DH_INTERNAL_OPTIONS}) ||
 	    grep /^-/, @ARGV) {
-		eval "use Debian::Debhelper::Dh_Getopt";
+		eval { require Debian::Debhelper::Dh_Getopt; };
 		error($@) if $@;
 		Debian::Debhelper::Dh_Getopt::parseopts(%params);
 	}
@@ -172,14 +178,20 @@ sub init {
 	# If no error handling function was specified, just propagate
 	# errors out.
 	if (! exists $dh{ERROR_HANDLER} || ! defined $dh{ERROR_HANDLER}) {
-		$dh{ERROR_HANDLER}='exit \$?';
+		$dh{ERROR_HANDLER}='exit 1';
 	}
+
+	$dh{U_PARAMS} //= [];
 }
 
 # Run at exit. Add the command to the log files for the packages it acted
 # on, if it's exiting successfully.
 my $write_log=1;
 sub END {
+	# If there is no 'debian/control', then we are not being run from
+	# a package directory and then the write_log will not do what we
+	# expect.
+	return if not -f 'debian/control';
 	if ($? == 0 && $write_log && (compat(9, 1) || $ENV{DH_INTERNAL_OVERRIDE})) {
 		write_log(basename($0), @{$dh{DOPACKAGES}});
 	}
@@ -310,11 +322,10 @@ sub print_and_doit_noerror {
 sub _doit {
 	my (@cmd) = @_;
 	my $options = ref($cmd[0]) ? shift(@cmd) : undef;
-	# In compat <= 10, we warn, compat 11 we detect and error, in
-	# compat 12 we assume people know what they are doing.
-	if (not defined($options) and @cmd == 1 and compat(11) and $cmd[0] =~ m/[\s<&>|;]/) {
-		deprecated_functionality('doit() + doit_*() calls will no longer spawn a shell in compat 11 for single string arguments (please use complex_doit instead)',
-								 11);
+	# In compat <= 11, we warn, in compat 12 we assume people know what they are doing.
+	if (not defined($options) and @cmd == 1 and compat(12) and $cmd[0] =~ m/[\s<&>|;]/) {
+		deprecated_functionality('doit() + doit_*() calls will no longer spawn a shell in compat 12 for single string arguments (please use complex_doit instead)',
+								 12);
 		return 1 if $dh{NO_ACT};
 		return system(@cmd) == 0;
 	}
@@ -322,9 +333,23 @@ sub _doit {
 	my $pid = fork() // error("fork(): $!");
 	if (not $pid) {
 		if (defined($options)) {
+			if (defined(my $dir = $options->{chdir})) {
+				if ($dir ne '.') {
+					chdir($dir) or error("chdir(\"${dir}\) failed: $!");
+				}
+			}
 			if (defined(my $output = $options->{stdout})) {
 				open(STDOUT, '>', $output) or error("redirect STDOUT failed: $!");
 			}
+			if (defined(my $update_env = $options->{update_env})) {
+				while (my ($k, $v) = each(%{$update_env})) {
+					if (defined($v)) {
+						$ENV{$k} = $v;
+					} else {
+						delete($ENV{$k});
+					}
+				}
+			}
 		}
 		# Force execvp call to avoid shell.  Apparently, even exec can
 		# involve a shell if you don't do this.
@@ -337,6 +362,28 @@ sub _format_cmdline {
 	my (@cmd) = @_;
 	my $options = ref($cmd[0]) ? shift(@cmd) : {};
 	my $cmd_line = escape_shell(@cmd);
+	if (defined(my $update_env = $options->{update_env})) {
+		my $need_env = 0;
+		my @params;
+		for my $key (sort(keys(%{$update_env}))) {
+			my $value = $update_env->{$key};
+			if (defined($value)) {
+				my $quoted_key = escape_shell($key);
+				push(@params, join('=', $quoted_key, escape_shell($value)));
+				# shell does not like: "FU BAR"=1 cmd
+				# if the ENV key has weird symbols, the best bet is to use env
+				$need_env = 1 if $quoted_key ne $key;
+			} else {
+				$need_env = 1;
+				push(@params, escape_shell("--unset=${key}"));
+			}
+		}
+		unshift(@params, 'env', '--') if $need_env;
+		$cmd_line = join(' ', @params, $cmd_line);
+	}
+	if (defined(my $dir = $options->{chdir})) {
+		$cmd_line = join(' ', 'cd', escape_shell($dir), '&&', $cmd_line) if $dir ne '.';
+	}
 	if (defined(my $output = $options->{stdout})) {
 		$cmd_line .= ' > ' . escape_shell($output);
 	}
@@ -497,6 +544,7 @@ sub rename_path {
 sub reset_perm_and_owner {
 	my ($mode, @paths) = @_;
 	my $_mode;
+	my $use_root = should_use_root();
 	# Dark goat blood to tell 0755 from "0755"
 	if (length( do { no warnings "numeric"; $mode & "" } ) ) {
 		# 0755, leave it alone.
@@ -507,12 +555,14 @@ sub reset_perm_and_owner {
 	}
 	if ($dh{VERBOSE}) {
 		verbose_print(sprintf('chmod %#o -- %s', $_mode, escape_shell(@paths)));
-		verbose_print(sprintf('chown 0:0 -- %s', escape_shell(@paths)));
+		verbose_print(sprintf('chown 0:0 -- %s', escape_shell(@paths))) if $use_root;
 	}
 	return if $dh{NO_ACT};
 	for my $path (@paths) {
 		chmod($_mode, $path) or error(sprintf('chmod(%#o, %s): %s', $mode, $path, $!));
-		chown(0, 0, $path) or error("chown(0, 0, $path): $!");
+		if ($use_root) {
+			chown(0, 0, $path) or error("chown(0, 0, $path): $!");
+		}
 	}
 }
 
@@ -628,6 +678,7 @@ sub dirname {
 
 # Pass in a number, will return true iff the current compatibility level
 # is less than or equal to that number.
+my $compat_from_bd;
 {
 	my $warned_compat = $ENV{DH_INTERNAL_TESTSUITE_SILENT_WARNINGS} ? 1 : 0;
 	my $c;
@@ -635,6 +686,8 @@ sub dirname {
 	sub compat {
 		my $num=shift;
 		my $nowarn=shift;
+
+		getpackages() if not defined($compat_from_bd);
 	
 		if (! defined $c) {
 			$c=1;
@@ -648,17 +701,23 @@ sub dirname {
 				}
 				else {
 					chomp $l;
-					$c=$l;
-					$c =~ s/^\s*+//;
-					$c =~ s/\s*+$//;
-					if (exists($NAMED_COMPAT_LEVELS{$c})) {
-						$c = $NAMED_COMPAT_LEVELS{$c};
-					} elsif ($c !~ m/^\d+$/) {
-						error("debian/compat must contain a positive number (found: \"$c\")");
+					my $new_compat = $l;
+					$new_compat =~ s/^\s*+//;
+					$new_compat =~ s/\s*+$//;
+					if ($new_compat !~ m/^\d+$/) {
+						error("debian/compat must contain a positive number (found: \"${new_compat}\")");
 					}
+					if (defined($compat_from_bd) and $compat_from_bd != -1) {
+						warning("Please specific the debhelper compat level exactly once.");
+						warning(" * debian/compat requests compat ${new_compat}.");
+						warning(" * debian/control requests compat ${compat_from_bd} via \"debhelper-compat (= ${compat_from_bd})\"");
+						error("debhelper compat level specified both in debian/compat and via build-dependency on debhelper-compat");
+					}
+					$c = $new_compat;
 				}
-			}
-			elsif (not $nowarn) {
+			} elsif ($compat_from_bd != -1) {
+				$c = $compat_from_bd;
+			} elsif (not $nowarn) {
 				error("Please specify the compatibility level in debian/compat");
 			}
 
@@ -710,14 +769,14 @@ sub default_sourcedir {
 # for the package, and it will return the actual existing filename to use.
 #
 # It tries several filenames:
-#   * debian/package.filename.buildarch
-#   * debian/package.filename.buildos
+#   * debian/package.filename.hostarch
+#   * debian/package.filename.hostos
 #   * debian/package.filename
 #   * debian/filename (if the package is the main package)
 # If --name was specified then the files
 # must have the name after the package name:
-#   * debian/package.name.filename.buildarch
-#   * debian/package.name.filename.buildos
+#   * debian/package.name.filename.hostarch
+#   * debian/package.name.filename.hostos
 #   * debian/package.name.filename
 #   * debian/name.filename (if the package is the main package)
 
@@ -762,19 +821,21 @@ sub default_sourcedir {
 			for my $pkg (@{$package}) {
 				push(@try, "debian/${pkg}.${filename}");
 				if ($check_expensive) {
+					my $cross_type = uc(package_cross_type($pkg));
 					push(@try,
-						 "debian/${pkg}.${filename}.".buildarch(),
-						 "debian/${pkg}.${filename}.".buildos(),
+						 "debian/${pkg}.${filename}.".dpkg_architecture_value("DEB_${cross_type}_ARCH"),
+						 "debian/${pkg}.${filename}.".dpkg_architecture_value("DEB_${cross_type}_ARCH_OS"),
 					);
 				}
 			}
 		} else {
-			# Avoid checking for buildarch+buildos unless we have reason
+			# Avoid checking for hostarch+hostos unless we have reason
 			# to believe that they exist.
 			if ($check_expensive) {
+				my $cross_type = uc(package_cross_type($package));
 				push(@try,
-					 "debian/${package}.${filename}.".buildarch(),
-					 "debian/${package}.${filename}.".buildos(),
+					 "debian/${package}.${filename}.".dpkg_architecture_value("DEB_${cross_type}_ARCH"),
+					 "debian/${package}.${filename}.".dpkg_architecture_value("DEB_${cross_type}_ARCH_OS"),
 					);
 			}
 			push(@try, "debian/$package.$filename");
@@ -829,19 +890,30 @@ sub pkgfilename {
 		my $package=shift;
 
 		return $isnative_cache{$package} if defined $isnative_cache{$package};
-		
+
+		if (not %isnative_cache) {
+			require Dpkg::Changelog::Parse;
+		}
+
 		# Make sure we look at the correct changelog.
 		my $isnative_changelog=pkgfile($package,"changelog");
 		if (! $isnative_changelog) {
 			$isnative_changelog="debian/changelog";
 		}
-		# Get the package version.
-		my $version=`dpkg-parsechangelog -l$isnative_changelog -SVersion`;
-		chomp($dh{VERSION} = $version);
-		# Did the changelog parse fail?
-		if ($dh{VERSION} eq q{}) {
-			error("changelog parse failure");
+		my $res = Dpkg::Changelog::Parse::changelog_parse(
+			file => $isnative_changelog,
+			compression => 0,
+		);
+		if (not defined($res)) {
+			error("No changelog entries for $package!? (changelog file: ${isnative_changelog})");
 		}
+		my $version = $res->{'Version'};
+		# Do we have a valid version?
+		if (not defined($version) or not $version->is_valid) {
+			error("changelog parse failure; invalid or missing version");
+		}
+		# Get the package version.
+		$dh{VERSION} = $version->as_string;
 
 		# Is this a native Debian package?
 		if (index($dh{VERSION}, '-') > -1) {
@@ -880,12 +952,20 @@ sub _tool_version {
 # 3: filename of snippet
 # 4: either text: shell-quoted sed to run on the snippet. Ie, 's/#PACKAGE#/$PACKAGE/'
 #    or a sub to run on each line of the snippet. Ie sub { s/#PACKAGE#/$PACKAGE/ }
+#    or a hashref with keys being variables and values being their replacement.  Ie. { PACKAGE => $PACKAGE }
+# 5: Internal usage only
 sub autoscript {
-	my ($package, $script, $filename, $sed) = @_;
+	my ($package, $script, $filename, $sed, $extra_options) = @_;
 
 	my $tool_version = _tool_version();
 	# This is the file we will modify.
 	my $outfile="debian/".pkgext($package)."$script.debhelper";
+	if ($extra_options && exists($extra_options->{'snippet-order'})) {
+		my $order = $extra_options->{'snippet-order'};
+		error("Internal error - snippet order set to unknown value: \"${order}\"")
+			if $order ne 'service';
+		$outfile = generated_file($package, "${script}.${order}");
+	}
 
 	# Figure out what shell script snippet to use.
 	my $infile;
@@ -907,16 +987,18 @@ sub autoscript {
 		# Add fragments to top so they run in reverse order when removing.
 		if (not defined($sed) or ref($sed)) {
 			verbose_print("[META] Prepend autosnippet \"$filename\" to $script [${outfile}.new]");
-			open(my $out_fd, '>', "${outfile}.new") or error("open(${outfile}.new): $!");
-			print {$out_fd} '# Automatically added by ' . basename($0) . "/${tool_version}\n";
-			autoscript_sed($sed, $infile, undef, $out_fd);
-			print {$out_fd} "# End automatically added section\n";
-			open(my $in_fd, '<', $outfile) or error("open($outfile): $!");
-			while (my $line = <$in_fd>) {
-				print {$out_fd} $line;
+			if (not $dh{NO_ACT}) {
+				open(my $out_fd, '>', "${outfile}.new") or error("open(${outfile}.new): $!");
+				print {$out_fd} '# Automatically added by ' . basename($0) . "/${tool_version}\n";
+				autoscript_sed($sed, $infile, undef, $out_fd);
+				print {$out_fd} "# End automatically added section\n";
+				open(my $in_fd, '<', $outfile) or error("open($outfile): $!");
+				while (my $line = <$in_fd>) {
+					print {$out_fd} $line;
+				}
+				close($in_fd);
+				close($out_fd) or error("close(${outfile}.new): $!");
 			}
-			close($in_fd);
-			close($out_fd) or error("close(${outfile}.new): $!");
 		} else {
 			complex_doit("echo \"# Automatically added by ".basename($0)."/${tool_version}\"> $outfile.new");
 			autoscript_sed($sed, $infile, "$outfile.new");
@@ -926,11 +1008,13 @@ sub autoscript {
 		rename_path("${outfile}.new", $outfile);
 	} elsif (not defined($sed) or ref($sed)) {
 		verbose_print("[META] Append autosnippet \"$filename\" to $script [${outfile}]");
-		open(my $out_fd, '>>', $outfile) or error("open(${outfile}): $!");
-		print {$out_fd} '# Automatically added by ' . basename($0) . "/${tool_version}\n";
-		autoscript_sed($sed, $infile, undef, $out_fd);
-		print {$out_fd} "# End automatically added section\n";
-		close($out_fd) or error("close(${outfile}): $!");
+		if (not $dh{NO_ACT}) {
+			open(my $out_fd, '>>', $outfile) or error("open(${outfile}): $!");
+			print {$out_fd} '# Automatically added by ' . basename($0) . "/${tool_version}\n";
+			autoscript_sed($sed, $infile, undef, $out_fd);
+			print {$out_fd} "# End automatically added section\n";
+			close($out_fd) or error("close(${outfile}): $!");
+		}
 	} else {
 		complex_doit("echo \"# Automatically added by ".basename($0)."/${tool_version}\">> $outfile");
 		autoscript_sed($sed, $infile, $outfile);
@@ -1010,6 +1094,9 @@ sub autoscript_sed {
 }
 
 # Generated files are cleaned by dh_clean AND dh_prep
+# - Package can be set to "_source" to generate a file relevant
+#   for the source package (the meson build does this atm.).
+#   Files for "_source" are only cleaned by dh_clean.
 sub generated_file {
 	my ($package, $filename, $mkdirs) = @_;
 	my $dir = "debian/.debhelper/generated/${package}";
@@ -1112,8 +1199,14 @@ sub glob_expand {
 	for my $pattern (@patterns) {
 		my @m;
 		for my $dir (@dirs) {
-			@m = bsd_glob("$dir/$pattern", GLOB_CSH & ~(GLOB_NOMAGIC|GLOB_TILDE));
-			last if @m;# > 1 or (@m and (-l $m[0] or -e _));
+			my $full_pattern = "$dir/$pattern";
+			@m = bsd_glob($full_pattern, GLOB_CSH & ~(GLOB_NOMAGIC|GLOB_TILDE));
+			last if @m;
+			# Handle "foo{bar}" pattern (#888251)
+			if (-l $full_pattern or -e _) {
+				push(@m, $full_pattern);
+				last;
+			}
 		}
 		if (not @m) {
 			$error_handler //= \&glob_expand_error_handler_reject;
@@ -1230,14 +1323,15 @@ sub excludefile {
 	}
 }
 
-# Returns the build architecture.
+# Confusing name for hostarch
 sub buildarch {
-	dpkg_architecture_value('DEB_HOST_ARCH');
+	deprecated_functionality('buildarch() is deprecated and replaced by hostarch()', 12);
+	goto \&hostarch;
 }
 
-# Returns the build OS.
-sub buildos {
-	dpkg_architecture_value("DEB_HOST_ARCH_OS");
+# Returns the architecture that will run binaries produced (DEB_HOST_ARCH)
+sub hostarch {
+	dpkg_architecture_value('DEB_HOST_ARCH');
 }
 
 # Returns a truth value if this seems to be a cross-compile
@@ -1273,33 +1367,25 @@ sub is_cross_compiling {
 	}
 }
 
-# Returns source package name
-sub sourcepackage {
-	open (my $fd, '<', 'debian/control') ||
-	    error("cannot read debian/control: $!\n");
-	while (<$fd>) {
-		chomp;
-		s/\s+$//;
-		if (/^Source:\s*(.*)/i) {
-			close($fd);
-			return $1;
-		}
-	}
 
-	close($fd);
-	error("could not find Source: line in control file.");
-}
 
 # Returns a list of packages in the control file.
-# Pass "arch" or "indep" to specify arch-dependant (that will be built
-# for the system's arch) or independant. If nothing is specified,
+# Pass "arch" or "indep" to specify arch-dependent (that will be built
+# for the system's arch) or independent. If nothing is specified,
 # returns all packages. Also, "both" returns the union of "arch" and "indep"
 # packages.
 #
 # As a side effect, populates %package_arches and %package_types
 # with the types of all packages (not only those returned).
 my (%package_types, %package_arches, %package_multiarches, %packages_by_type,
-    %package_sections);
+    %package_sections, $sourcepackage, %package_cross_type);
+
+# Returns source package name
+sub sourcepackage {
+	getpackages() if not defined($sourcepackage);
+	return $sourcepackage;
+}
+
 sub getpackages {
 	my ($type) = @_;
 	error("getpackages: First argument must be one of \"arch\", \"indep\", or \"both\"")
@@ -1316,7 +1402,8 @@ sub getpackages {
 	my $arch="";
 	my $section="";
 	my ($package_type, $multiarch, %seen, @profiles, $source_section,
-		$included_in_build_profile);
+		$included_in_build_profile, $cross_type, $cross_target_arch,
+		%bd_fields, $bd_field_value);
 	if (exists $ENV{'DEB_BUILD_PROFILES'}) {
 		@profiles=split /\s+/, $ENV{'DEB_BUILD_PROFILES'};
 	}
@@ -1327,6 +1414,80 @@ sub getpackages {
 	};
 
 	$packages_by_type{$_} = [] for qw(both indep arch all-listed-in-control-file);
+	while (<$fd>) {
+		chomp;
+		s/\s+$//;
+		next if m/^\s*+\#/;
+
+		if (/^Source:\s*(.*)/i) {
+			$sourcepackage = $1;
+			$bd_field_value = undef;
+			next;
+		} elsif (/^Section:\s(.*)$/i) {
+			$source_section = $1;
+			$bd_field_value = undef;
+			next;
+		} elsif (/^(Build-Depends(?:-Arch|-Indep)?):\s*(.*)$/i) {
+			my ($field, $value) = (lc($1), $2);
+			$bd_field_value = [$value];
+			$bd_fields{$field} = $bd_field_value;
+		} elsif (/^\S/) {
+			$bd_field_value = undef;
+		} elsif (/^\s/ and $bd_field_value) {
+			push(@{$bd_field_value}, $_);
+		}
+		next if not $_ and not defined($sourcepackage);
+		last if (!$_ or eof); # end of stanza.
+	}
+	error("could not find Source: line in control file.") if not defined($sourcepackage);
+	if (%bd_fields) {
+		my ($dh_compat_bd, $final_level);
+		for my $field (sort(keys(%bd_fields))) {
+			my $value = join(' ', @{$bd_fields{$field}});
+			$value =~ s/\s*,\s*$//;
+			for my $dep (split(/\s*,\s*/, $value)) {
+				if ($dep =~ m/^debhelper-compat\s*[(]\s*=\s*(${PKGVERSION_REGEX})\s*[)]$/) {
+					my $version = $1;
+					if ($version =~m/^(\d+)\D.*$/) {
+						my $guessed_compat = $1;
+						warning("Please use the compat level as the exact version rather than the full version.");
+						warning("  Perhaps you meant: debhelper-compat (= ${guessed_compat})");
+						if ($field ne 'build-depends') {
+							warning(" * Also, please move the declaration to Build-Depends (it was found in ${field})");
+						}
+						error("Invalid compat level ${version}, derived from relation: ${dep}");
+					}
+					$final_level = $version;
+					error("Duplicate debhelper-compat build-dependency: ${dh_compat_bd} vs. ${dep}") if $dh_compat_bd;
+					error("The debhelper-compat build-dependency must be in the Build-Depends field (not $field)")
+						if $field ne 'build-depends';
+					$dh_compat_bd = $dep;
+				} elsif ($dep =~ m/^debhelper-compat\s*(?:\S.*)?$/) {
+					my $clevel = "${\MAX_COMPAT_LEVEL}";
+					eval {
+						require Debian::Debhelper::Dh_Version;
+						$clevel = $Debian::Debhelper::Dh_Version::version;
+					};
+					$clevel =~ s/^\d+\K\D.*$//;
+					warning("Found invalid debhelper-compat relation: ${dep}");
+					warning(" * Please format the relation as (example): debhelper-compat (= ${clevel})");
+					warning(" * Note that alternatives, architecture restrictions, build-profiles etc. are not supported.");
+					if ($field ne 'build-depends') {
+						warning(" * Also, please move the declaration to Build-Depends (it was found in ${field})");
+					}
+					warning(" * If this is not possible, then please remove the debhelper-compat relation and insert the");
+					warning("   compat level into the file debian/compat.  (E.g. \"echo ${clevel} > debian/compat\")");
+					error("Could not parse desired debhelper compat level from relation: $dep");
+				}
+			}
+		}
+		if (defined($final_level)) {
+			warning("The use of \"debhelper-compat (= ${final_level})\" is experimental and may change (or be retired) without notice");
+		}
+		$compat_from_bd = $final_level // -1;
+	} else {
+		$compat_from_bd = -1;
+	}
 
 	while (<$fd>) {
 		chomp;
@@ -1349,7 +1510,11 @@ sub getpackages {
 			$package_type=$1;
 		} elsif (/^Multi-Arch:\s*(.*)/i) {
 			$multiarch = $1;
-
+		} elsif (/^X-DH-Build-For-Type:\s*(.*)/i) {
+			$cross_type = $1;
+			if ($cross_type ne 'host' and $cross_type ne 'target') {
+				error("Unknown value of X-DH-Build-For-Type \"$cross_type\" at debian/control:$.");
+			}
 		} elsif (/^Build-Profiles:\s*(.*)/i) {
 			# rely on libdpkg-perl providing the parsing functions
 			# because if we work on a package with a Build-Profiles
@@ -1374,22 +1539,34 @@ sub getpackages {
 				$package_arches{$package}=$arch;
 				$package_multiarches{$package} = $multiarch;
 				$package_sections{$package} = $section || $source_section;
+				$cross_type //= 'host';
+				$package_cross_type{$package} = $cross_type;
 				push(@{$packages_by_type{'all-listed-in-control-file'}}, $package);
 				if ($included_in_build_profile) {
 					if ($arch eq 'all') {
 						push(@{$packages_by_type{'indep'}}, $package);
 						push(@{$packages_by_type{'both'}}, $package);
-					} elsif ($arch eq 'any' ||
-							 ($arch ne 'all' && samearch(buildarch(), $arch))) {
-						push(@{$packages_by_type{'arch'}}, $package);
-						push(@{$packages_by_type{'both'}}, $package);
+					} else {
+						my $included = 0;
+						$included = 1 if $arch eq 'any';
+						if (not $included) {
+							my $desired_arch = hostarch();
+							if ($cross_type eq 'target') {
+								$cross_target_arch //= dpkg_architecture_value('DEB_TARGET_ARCH');
+								$desired_arch = $cross_target_arch;
+							}
+							$included = 1 if samearch($desired_arch, $arch);
+						}
+						if ($included) {
+							push(@{$packages_by_type{'arch'}}, $package);
+							push(@{$packages_by_type{'both'}}, $package);
+						}
 					}
 				}
-			} elsif ($section and not defined($source_section)) {
-				$source_section = $section;
 			}
 			$package='';
 			$package_type=undef;
+			$cross_type = undef;
 			$arch='';
 			$section='';
 		}
@@ -1399,6 +1576,43 @@ sub getpackages {
 	return @{$packages_by_type{$type}};
 }
 
+# Return true if we should use root.
+# - Takes an optional keyword; if passed, this will return true if the keyword is listed in R^3 (Rules-Requires-Root)
+# - If the optional keyword is omitted or not present in R^3 and R^3 is not 'binary-targets', then returns false
+# - Returns true otherwise (i.e. keyword is in R^3 or R^3 is 'binary-targets')
+{
+	my %rrr;
+	sub should_use_root {
+		my ($keyword) = @_;
+		my $rrr_env = $ENV{'DEB_RULES_REQUIRES_ROOT'} // 'binary-targets';
+		$rrr_env =~ s/^\s++//;
+		$rrr_env =~ s/\s++$//;
+		return 0 if $rrr_env eq 'no';
+		return 1 if $rrr_env eq 'binary-targets';
+		return 0 if not defined($keyword);
+
+		%rrr = map { $_ => 1 } split(' ', $rrr_env) if not %rrr;
+		return 1 if exists($rrr{$keyword});
+		return 0;
+	}
+}
+
+# Returns the "gain root command" as a list suitable for passing as a part of the command to "doit()"
+sub gain_root_cmd {
+	my $raw_cmd = $ENV{DEB_GAIN_ROOT_CMD};
+	return if not defined($raw_cmd) or $raw_cmd =~ m/^\s*+$/;
+	return split(' ', $raw_cmd);
+}
+
+sub root_requirements {
+	my $rrr_env = $ENV{'DEB_RULES_REQUIRES_ROOT'} // 'binary-targets';
+	$rrr_env =~ s/^\s++//;
+	$rrr_env =~ s/\s++$//;
+	return 'none' if $rrr_env eq 'no';
+	return 'legacy-root' if $rrr_env eq 'binary-targets';
+	return 'targeted-promotion';
+}
+
 # Returns the arch a package will build for.
 #
 # Deprecated: please switch to the more descriptive
@@ -1415,9 +1629,11 @@ sub package_binary_arch {
 
 	if (! exists $package_arches{$package}) {
 		warning "package $package is not in control info";
-		return buildarch();
+		return hostarch();
 	}
-	return $package_arches{$package} eq 'all' ? "all" : buildarch();
+	return 'all' if $package_arches{$package} eq 'all';
+	return dpkg_architecture_value('DEB_TARGET_ARCH') if package_cross_type($package) eq 'target';
+	return hostarch();
 }
 
 # Returns the Architecture: value which the package declared.
@@ -1426,7 +1642,7 @@ sub package_declared_arch {
 
 	if (! exists $package_arches{$package}) {
 		warning "package $package is not in control info";
-		return buildarch();
+		return hostarch();
 	}
 	return $package_arches{$package};
 }
@@ -1437,7 +1653,7 @@ sub package_is_arch_all {
 
 	if (! exists $package_arches{$package}) {
 		warning "package $package is not in control info";
-		return buildarch();
+		return hostarch();
 	}
 	return $package_arches{$package} eq 'all';
 }
@@ -1469,6 +1685,18 @@ sub package_section {
 	return $package_sections{$package} // 'unknown';
 }
 
+sub package_cross_type {
+	my ($package) = @_;
+
+	# Test the architecture field instead, as it is common for a
+	# package to not have a multi-arch value.
+	if (! exists $package_cross_type{$package}) {
+		warning "package $package is not in control info";
+		return 'host';
+	}
+	return $package_cross_type{$package} // 'host';
+}
+
 # Return true if a given package is really a udeb.
 sub is_udeb {
 	my $package=shift;
@@ -1492,6 +1720,18 @@ sub is_udeb {
 	}
 }
 
+sub _concat_slurp_script_files {
+	my (@files) = @_;
+	my $res = '';
+	for my $file (@files) {
+		open(my $fd, '<', $file) or error("open($file) failed: $!");
+		my $f = join('', <$fd>);
+		close($fd);
+		$res .= $f;
+	}
+	return $res;
+}
+
 # Handles #DEBHELPER# substitution in a script; also can generate a new
 # script from scratch if none exists but there is a .debhelper file for it.
 sub debhelper_script_subst {
@@ -1501,34 +1741,56 @@ sub debhelper_script_subst {
 	my $tmp=tmpdir($package);
 	my $ext=pkgext($package);
 	my $file=pkgfile($package,$script);
+	my $service_script = generated_file($package, "${script}.service", 0);
+	my @generated_scripts = ("debian/$ext$script.debhelper", $service_script);
+	if ($script eq 'prerm' or $script eq 'postrm') {
+		@generated_scripts = reverse(@generated_scripts);
+	}
+	@generated_scripts = grep { -f } @generated_scripts;
 
 	if ($file ne '') {
-		if (-f "debian/$ext$script.debhelper") {
+		if (@generated_scripts) {
+			if ($dh{VERBOSE}) {
+				verbose_print('cp -f ' . escape_shell($file) . " $tmp/DEBIAN/$script");
+				verbose_print("perl -p -i -e \"s~#DEBHELPER#~qx{cat @generated_scripts}~eg\" $tmp/DEBIAN/$script");
+			}
 			# Add this into the script, where it has #DEBHELPER#
-			doit({ stdout => "$tmp/DEBIAN/$script" }, 'perl', '-pe',
-				 "s~#DEBHELPER#~qx{cat debian/$ext$script.debhelper}~eg", $file);
-		}
-		else {
+			my $text = _concat_slurp_script_files(@generated_scripts);
+			if (not $dh{NO_ACT}) {
+				open(my $out_fd, '>', "$tmp/DEBIAN/$script") or error("open($tmp/DEBIAN/$script) failed: $!");
+				open(my $in_fd, '<', $file) or error("open($file) failed: $!");
+				while (my $line = <$in_fd>) {
+					$line =~ s/#DEBHELPER#/$text/g;
+					print {$out_fd} $line;
+				}
+				close($in_fd);
+				close($out_fd) or error("close($tmp/DEBIAN/$script) failed: $!");
+			}
+		} else {
 			# Just get rid of any #DEBHELPER# in the script.
 			doit({ stdout => "$tmp/DEBIAN/$script" }, 'sed', 's/#DEBHELPER#//', $file);
 		}
 		reset_perm_and_owner('0755', "$tmp/DEBIAN/$script");
 	}
-	elsif ( -f "debian/$ext$script.debhelper" ) {
+	elsif (@generated_scripts) {
 		if ($dh{VERBOSE}) {
 			verbose_print(q{printf '#!/bin/sh\nset -e\n' > } . "$tmp/DEBIAN/$script");
-			verbose_print("cat debian/$ext$script.debhelper >> $tmp/DEBIAN/$script");
+			verbose_print("cat @generated_scripts >> $tmp/DEBIAN/$script");
 		}
-		open(my $out_fd, '>', "$tmp/DEBIAN/$script") or error("open($tmp/DEBIAN/$script): $!");
-		print {$out_fd} "#!/bin/sh\n";
-		print {$out_fd} "set -e\n";
-		open(my $in_fd, '<', "debian/$ext$script.debhelper")
-			or error("open(debian/$ext$script.debhelper): $!");
-		while (my $line = <$in_fd>) {
-			print {$out_fd} $line;
+		if (not $dh{NO_ACT}) {
+			open(my $out_fd, '>', "$tmp/DEBIAN/$script") or error("open($tmp/DEBIAN/$script): $!");
+			print {$out_fd} "#!/bin/sh\n";
+			print {$out_fd} "set -e\n";
+			for my $generated_script (@generated_scripts) {
+				open(my $in_fd, '<', $generated_script)
+					or error("open($generated_script) failed: $!");
+				while (my $line = <$in_fd>) {
+					print {$out_fd} $line;
+				}
+				close($in_fd);
+			}
+			close($out_fd) or error("close($tmp/DEBIAN/$script) failed: $!");
 		}
-		close($in_fd);
-		close($out_fd) or error("close($tmp/DEBIAN/$script): $!");
 		reset_perm_and_owner('0755', "$tmp/DEBIAN/$script");
 	}
 }
@@ -1564,9 +1826,17 @@ sub make_symlink{
 	my $dest = shift;
 	my $src = _expand_path(shift);
 	my $tmp = shift;
-        $tmp = '' if not defined($tmp);
-	$src=~s:^/::;
-	$dest=~s:^/::;
+	$tmp = '' if not defined($tmp);
+
+	if ($dest =~ m{(?:^|/)*[.]{2}(?:/|$)}) {
+		error("Invalid destination/link name (contains \"..\"-segments): $dest");
+	}
+
+	$src =~ s{^(?:[.]/+)++}{};
+	$dest =~ s{^(?:[.]/+)++}{};
+
+	$src=~s:^/++::;
+	$dest=~s:^/++::;
 
 	if ($src eq $dest) {
 		warning("skipping link from $src to self");
@@ -1578,8 +1848,8 @@ sub make_symlink{
 	# Policy says that if the link is all within one toplevel
 	# directory, it should be relative. If it's between
 	# top level directories, leave it absolute.
-	my @src_dirs=split(m:/+:,$src);
-	my @dest_dirs=split(m:/+:,$dest);
+	my @src_dirs = grep { $_ ne '.' } split(m:/+:,$src);
+	my @dest_dirs = grep { $_ ne '.' } split(m:/+:,$dest);
 	if (@src_dirs > 0 && $src_dirs[0] eq $dest_dirs[0]) {
 		# Figure out how much of a path $src and $dest
 		# share in common.
@@ -1681,7 +1951,12 @@ sub clean_jobserver_makeflags {
 
 # If cross-compiling, returns appropriate cross version of command.
 sub cross_command {
-	my $command=shift;
+	my ($package, $command) = @_;
+	if (package_cross_type($package) eq 'target') {
+		if (dpkg_architecture_value("DEB_HOST_GNU_TYPE") ne dpkg_architecture_value("DEB_TARGET_GNU_TYPE")) {
+			return dpkg_architecture_value("DEB_TARGET_GNU_TYPE") . "-$command";
+		}
+	}
 	if (is_cross_compiling()) {
 		return dpkg_architecture_value("DEB_HOST_GNU_TYPE")."-$command";
 	}
@@ -1695,12 +1970,12 @@ sub cross_command {
 # variable and returns the computed value.
 sub get_source_date_epoch {
 	return $ENV{SOURCE_DATE_EPOCH} if exists($ENV{SOURCE_DATE_EPOCH});
-	eval "use Dpkg::Changelog::Debian";
+	eval { require Dpkg::Changelog::Debian };
 	if ($@) {
 		warning "unable to set SOURCE_DATE_EPOCH: $@";
 		return;
 	}
-	eval "use Time::Piece";
+	eval { require Time::Piece };
 	if ($@) {
 		warning "unable to set SOURCE_DATE_EPOCH: $@";
 		return;
@@ -1732,7 +2007,7 @@ sub set_buildflags {
 	# rely on this [CVE-2016-1238]
 	$ENV{PERL_USE_UNSAFE_INC} = 1 if compat(10);
 
-	eval "use Dpkg::BuildFlags";
+	eval { require Dpkg::BuildFlags };
 	if ($@) {
 		warning "unable to load build flags: $@";
 		return;
@@ -1928,12 +2203,83 @@ sub log_installed_files {
 	return 1;
 }
 
+use constant {
+	# The ELF header is at least 0x32 bytes (32bit); any filer shorter than that is not an ELF file
+	ELF_MIN_LENGTH => 0x32,
+	ELF_MAGIC => "\x7FELF",
+	ELF_ENDIAN_LE => 0x01,
+	ELF_ENDIAN_BE => 0x02,
+	ELF_TYPE_EXECUTABLE => 0x0002,
+	ELF_TYPE_SHARED_OBJECT => 0x0003,
+};
+
+sub is_so_or_exec_elf_file {
+	my ($file) = @_;
+	open(my $fd, '<:raw', $file) or error("open $file: $!");
+	my $buflen = 0;
+	my ($buf, $endian);
+	while ($buflen < ELF_MIN_LENGTH) {
+		my $r = read($fd, $buf, ELF_MIN_LENGTH - $buflen, $buflen) // error("read ($file): $!");
+		last if $r == 0; # EOF
+		$buflen += $r
+	}
+	close($fd);
+	return 0 if $buflen < ELF_MIN_LENGTH;
+
+	return 0 if substr($buf, 0x00, 4) ne ELF_MAGIC;
+	$endian = unpack('c', substr($buf, 0x05, 1));
+	my ($long_format, $short_format);
+
+	if ($endian == ELF_ENDIAN_BE) {
+		$long_format = 'N';
+		$short_format = 'n';
+	} elsif ($endian == ELF_ENDIAN_LE) {
+		$long_format = 'V';
+		$short_format = 'v';
+	} else {
+		return 0;
+	}
+	my $elf_version = substr($buf, 0x14, 4);
+	my $elf_type = substr($buf, 0x10, 2);
+
+
+	return 0 if unpack($long_format, $elf_version) != 0x00000001;
+	my $elf_type_unpacked = unpack($short_format, $elf_type);
+	return 0 if $elf_type_unpacked != ELF_TYPE_EXECUTABLE and $elf_type_unpacked != ELF_TYPE_SHARED_OBJECT;
+	return 1;
+}
 
 sub on_pkgs_in_parallel(&) {
 	unshift(@_, $dh{DOPACKAGES});
 	goto \&on_items_in_parallel;
 }
 
+# Given a list of files, find all hardlinked files and return:
+# 1: a list of unique files (all files in the list are not hardlinked with any other file in that list)
+# 2: a map where the keys are names of hardlinks and the value points to the name selected as the file put in the
+#    list of unique files.
+#
+# This is can be used to relink hard links after modifying one of them.
+sub find_hardlinks {
+	my (@all_files) = @_;
+	my (%seen, %hardlinks, @unique_files);
+	for my $file (@all_files) {
+		my ($dev, $inode, undef, $nlink)=stat($file);
+		if (defined $nlink && $nlink > 1) {
+			if (! $seen{"$inode.$dev"}) {
+				$seen{"$inode.$dev"}=$file;
+				push(@unique_files, $file);
+			} else {
+				# This is a hardlink.
+				$hardlinks{$file}=$seen{"$inode.$dev"};
+			}
+		} else {
+			push(@unique_files, $file);
+		}
+	}
+	return (\@unique_files, \%hardlinks);
+}
+
 sub on_items_in_parallel {
 	my ($pkgs_ref, $code) = @_;
 	my @pkgs = @{$pkgs_ref};
@@ -2014,4 +2360,22 @@ sub compute_doc_main_package {
 	return;
 }
 
+sub dbgsym_tmpdir {
+	my ($package) = @_;
+	return "debian/.debhelper/${package}/dbgsym-root";
+}
+
+
+{
+	my %known_packages;
+	sub assert_opt_is_known_package {
+		my ($package, $method) = @_;
+		%known_packages = map { $_ => 1 } getpackages() if not %known_packages;
+		if (not exists($known_packages{$package})) {
+			error("Requested unknown package $package via $method, expected one of: " . join(' ', getpackages()));
+		}
+		return 1;
+	}
+}
+
 1
